from random import randint as rint
from random import gauss
from math   import ceil, floor
import defs, graphic, sprite, missile
import pygame

def new_character_animation_sheet():
    new_sheet = []
    for animation in xrange(defs.numstates):
        new_sheet.append([])
    return new_sheet

def new_animation(spritesheet, frames, frame_duration):
    frame_list = [spritesheet[f] for f in frames]
    return sprite.SpriteAnimationSequenceFixed(frame_list, frame_duration)

def create_animation(spritesheet, (frames, frame_duration)):
    frame_list = [spritesheet[f] for f in frames]
    return sprite.SpriteAnimationSequenceFixed(frame_list, frame_duration)    

class Character(object):
    def __init__(self, sprite, rect, hitpoints, speed):
        self.sprite    = sprite
        self.rect      = rect
        self.hitpoints = hitpoints
        self.speed     = speed
        
        self.set_state(defs.state_idle)
        self.sprite.set_animation(defs.state_idle)
        
        self.delete_flag = False
        
    def can_delete(self):
        return self.delete_flag
    
    def hit(self, missile):
        self.hitpoints -= missile.damage
        if self.hitpoints > 0:
            self.set_state(defs.state_get_hit)
            self.state_timer = self.sprite.current_animation.duration
        else:
            self.set_state(defs.state_dying)
            self.state_timer = self.sprite.current_animation.duration
    
    def draw(self, surface, shift=(0, 0)):
        self.sprite.draw(surface, self.rect.topleft, shift)
    
    def set_state(self, state):
        self.state = state
        self.sprite.set_animation(state)
    
    def think(self, keys, events, world):
        pass
    
    def prepare_shot(self, target):
        self.target     = target
        self.fire_timer = self.fire_delay
    
    def shoot(self, world):
        pass
        
    def update(self, delay, keys, events, world):
        self.sprite.update(delay)
        
        self.state_timer = max(0, self.state_timer - delay)
        self.fire_timer  = max(0, self.fire_timer  - delay)
        
        if self.state in [defs.state_walkleft, defs.state_walkleft2]:
            self.rect.x = int(floor(self.rect.x - self.speed * delay))
            
        elif self.state in [defs.state_walkright, defs.state_walkright2]:
            self.rect.x = int(ceil(self.rect.x + self.speed * delay))

class Player(Character):
    def __init__(self, location):
        rect = pygame.rect.Rect(location, defs.player_size)

        image = graphic.load_image(defs.player_spritesheet)
        ss    = sprite.SpriteSheet(image, defs.player_size)
        
        # Check here
        animations = new_character_animation_sheet()
        animations[defs.state_idle]       = create_animation(ss, defs.p_animations["idle"])
        animations[defs.state_walkright]  = create_animation(ss, defs.p_animations["right"])
        animations[defs.state_walkleft]   = create_animation(ss, defs.p_animations["left"])
        animations[defs.state_idle2]      = create_animation(ss, defs.p_animations["idle2"])
        animations[defs.state_walkright2] = create_animation(ss, defs.p_animations["right2"])
        animations[defs.state_walkleft2]  = create_animation(ss, defs.p_animations["left2"])
        animations[defs.state_catch]      = create_animation(ss, defs.p_animations["catch"])
        animations[defs.state_attack]     = create_animation(ss, defs.p_animations["attack"])
        animations[defs.state_get_hit]    = create_animation(ss, defs.p_animations["hit"])
        animations[defs.state_dying]      = create_animation(ss, defs.p_animations["dying"])
        
        sp = sprite.Sprite(animations)
        
        Character.__init__(self, sp, rect, defs.player_hitpoints, defs.player_speed)
        
        self.lives       = defs.player_lives
        self.state_timer = 0
        self.fire_timer  = 0
        self.fire_delay  = defs.player_fire_delay
        self.weapon      = None
        
    def think(self, keys, events, world):
        # Is the player in the middle of something?
        if self.state_timer:
            return
        
        # Is the player trying to attack?
        for e in events:
            
            # Is the player trying to catch something?
            if e.type == pygame.KEYDOWN:
                if not self.weapon:
                    if e.key in defs.key_attack:
                        self.set_state(defs.state_catch)
                        self.state_timer = self.sprite.current_animation.duration
                        return
                    
                    if e.key == pygame.K_k:
                        world.enemies = []
                        return
            
            if e.type == pygame.MOUSEBUTTONDOWN:
                if self.weapon:
                    self.set_state(defs.state_attack)
                    
                    (tx, ty), tz = pygame.mouse.get_pos(), defs.background_z_center
                    self.prepare_shot((tx, ty, tz))
                    self.state_timer = self.sprite.current_animation.duration
                    return
        
        # Is the player moving left?
        for k in defs.key_left:
            if keys[k]:
                if not self.weapon:
                    self.set_state(defs.state_walkleft)
                else:
                    self.set_state(defs.state_walkleft2)
                return
        
        # Is the player moving right?
        for k in defs.key_right:
            if keys[k]:
                if not self.weapon:
                    self.set_state(defs.state_walkright)
                else:
                    self.set_state(defs.state_walkright2)
                return
        
        if not self.weapon:
            self.set_state(defs.state_idle)
        else:
            self.set_state(defs.state_idle2)
    
    def shoot(self, world):
        (sx, sy) = self.rect.center
        if self.weapon.type == "arrow":
            arrow = missile.Arrow((sx, sy, defs.foreground_z_center), self.target)
            world.add_missile(arrow)
        
        self.target = None
        self.weapon = None
    
    def update(self, delay, keys, events, world):
        Character.update(self, delay, keys, events, world)
        
        if self.state == defs.state_attack:
            if self.target:
                if self.fire_timer == 0:
                    self.shoot(world)
        
        if self.state == defs.state_dying:
            if self.state_timer == 0:
                self.hitpoints    = defs.player_hitpoints
                self.rect.centerx = 1024/2
                self.lives -= 1

class Archer(Character):
    def __init__(self, location):
        rect  = pygame.rect.Rect(location, defs.archer_size)
        image = graphic.load_image(defs.archer_spritesheet)
        ss    = sprite.SpriteSheet(image, defs.archer_size)
        
        animations = new_character_animation_sheet()
        animations[defs.state_arrive]    = create_animation(ss, defs.a_animations["arrive"])
        animations[defs.state_idle]      = create_animation(ss, defs.a_animations["idle"])
        animations[defs.state_walkright] = create_animation(ss, defs.a_animations["right"])
        animations[defs.state_walkleft]  = create_animation(ss, defs.a_animations["left"])
        animations[defs.state_attack]    = create_animation(ss, defs.a_animations["attack"])
        animations[defs.state_get_hit]   = create_animation(ss, defs.a_animations["hit"])
        animations[defs.state_dying]     = create_animation(ss, defs.a_animations["dying"])
        
        sp = sprite.Sprite(animations)
        
        Character.__init__(self, sp, rect, defs.archer_hitpoints, defs.archer_speed)
        
        self.state_timer   = 0
        self.fire_timer    = 0
        self.fire_delay    = defs.archer_fire_delay
        self.fire_cooldown = self.fire_delay
        self.fire_cooldown_max = defs.archer_fire_cooldown

    def think(self, keys, events, world):
        if self.state_timer:
            return
        if self.fire_cooldown == 0:
            self.set_state(defs.state_attack)
            self.state_timer   = self.sprite.current_animation.duration
            self.fire_timer    = self.fire_delay
            self.fire_cooldown = self.fire_cooldown_max
            return
        choice = rint(0, 2)
        if choice   == 0:
            self.set_state(defs.state_idle)
            self.state_timer = rint(500, 1000)
        elif choice == 1:
            self.set_state(defs.state_walkright)
            self.state_timer = rint(500, 1000)
        elif choice == 2:
            self.set_state(defs.state_walkleft)
            self.state_timer = rint(500, 1000)
    
    def update(self, delay, keys, events, world):
        Character.update(self, delay, keys, events, world)
        
        if self.rect.left <= defs.castle_x_range[0]:
            self.set_state(defs.state_walkright)
            self.state_timer = rint(1000, 2000)
            return
        
        if self.rect.right >= defs.castle_x_range[1]:
            self.set_state(defs.state_walkleft)
            self.state_timer = rint(1000, 2000)
            return
        
        self.fire_cooldown = max(0, self.fire_cooldown - delay)
        
        if self.state == defs.state_attack and self.state_timer == 0:
            self.state = defs.state_idle
            (sx, sy) = self.rect.center
            (tx, ty) = world.player.rect.center
            tx = gauss(tx, defs.archer_inaccuracy)
            arrow = missile.Arrow((sx, sy, defs.background_z_center), (tx, ty, defs.foreground_z_center))
            world.add_missile(arrow)
        
        if self.state == defs.state_dying and self.state_timer == 0:
            self.delete_flag = True

class StrongArcher(Archer):
    def __init__(self, location):
        rect  = pygame.rect.Rect(location, defs.archer_size)
        image = graphic.load_image(defs.archer_spritesheet)
        ss    = sprite.SpriteSheet(image, defs.archer_size)
        
        animations = new_character_animation_sheet()
        animations[defs.state_arrive]    = create_animation(ss, defs.sa_animations["arrive"])
        animations[defs.state_idle]      = create_animation(ss, defs.sa_animations["idle"])
        animations[defs.state_walkright] = create_animation(ss, defs.sa_animations["right"])
        animations[defs.state_walkleft]  = create_animation(ss, defs.sa_animations["left"])
        animations[defs.state_attack]    = create_animation(ss, defs.sa_animations["attack"])
        animations[defs.state_get_hit]   = create_animation(ss, defs.sa_animations["hit"])
        animations[defs.state_dying]     = create_animation(ss, defs.sa_animations["dying"])
        
        sp = sprite.Sprite(animations)
        
        Character.__init__(self, sp, rect, defs.s_archer_hitpoints, defs.s_archer_speed)
        
        self.state_timer   = 0
        self.fire_timer    = 0
        self.fire_delay    = defs.s_archer_fire_delay
        self.fire_cooldown = self.fire_delay
        self.fire_cooldown_max = defs.s_archer_fire_cooldown

class QuickArcher(Archer):
    def __init__(self, location):
        rect  = pygame.rect.Rect(location, defs.archer_size)
        image = graphic.load_image(defs.archer_spritesheet)
        ss    = sprite.SpriteSheet(image, defs.archer_size)
        
        animations = new_character_animation_sheet()
        animations[defs.state_arrive]    = create_animation(ss, defs.qa_animations["arrive"])
        animations[defs.state_idle]      = create_animation(ss, defs.qa_animations["idle"])
        animations[defs.state_walkright] = create_animation(ss, defs.qa_animations["right"])
        animations[defs.state_walkleft]  = create_animation(ss, defs.qa_animations["left"])
        animations[defs.state_attack]    = create_animation(ss, defs.qa_animations["attack"])
        animations[defs.state_get_hit]   = create_animation(ss, defs.qa_animations["hit"])
        animations[defs.state_dying]     = create_animation(ss, defs.qa_animations["dying"])
        
        sp = sprite.Sprite(animations)
        
        Character.__init__(self, sp, rect, defs.q_archer_hitpoints, defs.q_archer_speed)
        
        self.state_timer   = 0
        self.fire_timer    = 0
        self.fire_delay    = defs.q_archer_fire_delay
        self.fire_cooldown = self.fire_delay
        self.fire_cooldown_max = defs.q_archer_fire_cooldown

class EliteArcher(Archer):
    def __init__(self, location):
        rect  = pygame.rect.Rect(location, defs.archer_size)
        image = graphic.load_image(defs.archer_spritesheet)
        ss    = sprite.SpriteSheet(image, defs.archer_size)
        
        animations = new_character_animation_sheet()
        animations[defs.state_arrive]    = create_animation(ss, defs.ea_animations["arrive"])
        animations[defs.state_idle]      = create_animation(ss, defs.ea_animations["idle"])
        animations[defs.state_walkright] = create_animation(ss, defs.ea_animations["right"])
        animations[defs.state_walkleft]  = create_animation(ss, defs.ea_animations["left"])
        animations[defs.state_attack]    = create_animation(ss, defs.ea_animations["attack"])
        animations[defs.state_get_hit]   = create_animation(ss, defs.ea_animations["hit"])
        animations[defs.state_dying]     = create_animation(ss, defs.ea_animations["dying"])
        
        sp = sprite.Sprite(animations)
        
        Character.__init__(self, sp, rect, defs.e_archer_hitpoints, defs.e_archer_speed)
        
        self.state_timer   = 0
        self.fire_timer    = 0
        self.fire_delay    = defs.e_archer_fire_delay
        self.fire_cooldown = self.fire_delay
        self.fire_cooldown_max = defs.e_archer_fire_cooldown

class KingArcher(Archer):
    def __init__(self, location):
        rect  = pygame.rect.Rect(location, defs.archer_size)
        image = graphic.load_image(defs.archer_spritesheet)
        ss    = sprite.SpriteSheet(image, defs.archer_size)
        
        animations = new_character_animation_sheet()
        animations[defs.state_arrive]    = create_animation(ss, defs.ka_animations["arrive"])
        animations[defs.state_idle]      = create_animation(ss, defs.ka_animations["idle"])
        animations[defs.state_walkright] = create_animation(ss, defs.ka_animations["right"])
        animations[defs.state_walkleft]  = create_animation(ss, defs.ka_animations["left"])
        animations[defs.state_attack]    = create_animation(ss, defs.ka_animations["attack"])
        animations[defs.state_get_hit]   = create_animation(ss, defs.ka_animations["hit"])
        animations[defs.state_dying]     = create_animation(ss, defs.ka_animations["dying"])
        
        sp = sprite.Sprite(animations)
        
        Character.__init__(self, sp, rect, defs.k_archer_hitpoints, defs.k_archer_speed)
        
        self.state_timer   = 0
        self.fire_timer    = 0
        self.fire_delay    = defs.k_archer_fire_delay
        self.fire_cooldown = self.fire_delay
        self.fire_cooldown_max = defs.k_archer_fire_cooldown
